#!/usr/bin/env python
# coding: utf-8
from __future__ import print_function

import os
import sys
import time
import stat
import socket
import struct
import threading
import argparse

"""
sftp-srv.py: sftpd subsystem with access filtering
  (goes well with rsshfs.sh)
"""
__version__   = "0.2"
__date__      = "2020-06-02"
__author__    = "ed <irc.rizon.net>"
__url__       = "https://ocv.me/dev/?sftp-srv.py"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2020

"""
works:
- directory listing
- read regular files
not implemented:
- writing
- filtering
- symlinks
"""


PY2 = sys.version_info[0] == 2
if PY2:
    from cStringIO import StringIO as BytesIO
else:
    from io import BytesIO as BytesIO


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def b2hex(data):
    if PY2:
        return " ".join(map(lambda b: format(ord(b), "02x"), data))
    else:
        if type(data) is str:
            return " ".join(map(lambda b: format(ord(b), "02x"), data))
        else:
            return " ".join(map(lambda b: format(b, "02x"), data))


class PositionTracker(object):
    def __init__(self, buf):
        self.f = buf
        self.pos = 0

    def read(self, n=-1):
        rv = self.f.read(n)
        self.pos += len(rv)
        return rv

    def tell(self):
        return self.pos


class Parser(object):
    def __init__(self, buf):
        if hasattr(buf, "read"):
            self.sz = -1
            try:
                buf.tell()
                self.f = buf
                # eprint('[*] direct buffer access')
            except:
                self.f = PositionTracker(buf)
                # eprint('[*] indirect access using PositionTracker')
        else:
            # eprint('[*] indirect access using BytesIO')
            self.f = BytesIO(buf)
            self.sz = len(buf)

    def read(self, sz):
        ret = self.f.read(sz)
        while ret and len(ret) < sz:
            b = self.f.read(sz - len(ret))
            ret += b
            if not b:
                break
        
        if len(ret) == sz:
            return ret
        
        raise Exception("EOF")
    
    def u64(self):
        buf = self.read(8)
        return struct.unpack(b"<Q", buf)[0]

    def ru64(self):
        buf = self.read(8)
        return struct.unpack(b">Q", buf)[0]

    def s32(self):
        buf = self.read(4)
        return struct.unpack(b"<i", buf)[0]

    def u32(self):
        buf = self.read(4)
        return struct.unpack(b"<I", buf)[0]

    def ru32(self):
        buf = self.read(4)
        return struct.unpack(b">I", buf)[0]

    def u16(self):
        buf = self.read(2)
        return struct.unpack(b"<H", buf)[0]

    def ru16(self):
        buf = self.read(2)
        return struct.unpack(b">H", buf)[0]

    def u8(self):
        buf = self.read(1)
        return struct.unpack(b"B", buf)[0]

    def drop(self, n_bytes):
        self.read(n_bytes)

    def unk(self, expect):
        v = self.read(len(expect))
        if v != expect:
            raise Exception(
                "unexpected unknown:\n{0} got\n{1} expected".format(
                    b2hex(v), b2hex(expect)
                )
            )

    def _str(self, maxlen, sz):
        if sz > maxlen:
            raise Exception("string out of bounds: {0}".format(sz))

        return self.read(sz).decode("utf-8").rstrip("\0")
    
    def str32(self, maxlen):
        sz = self.u32()
        return self._str(maxlen, sz)

    def rstr32(self, maxlen):
        sz = self.ru32()
        return self._str(maxlen, sz)

    def align(self, to):
        pos = self.f.tell()
        waste = 0
        while divmod(pos, to)[1] != 0:
            waste += 1
            pos += 1

        if waste != 0:
            self.read(waste)


class FSocket(object):
    def __init__(self, sck):
        self.sck = sck

    def read(self, sz):
        return self.sck.recv(sz)

    def write(self, buf):
        nw = self.sck.send(buf)
        left = len(buf) - nw
        while left > 0:
            nw = self.sck.send(buf[:-left])
            left -= nw
            if nw == 0:
                raise Exception("client lost during write")


class FStdio(object):
    def __init__(self):
        if not PY2:
            self.stdin = sys.stdin.buffer
            self.stdout = sys.stdout.buffer
            self.write = self._write3
        else:
            if sys.platform == "win32":
                import msvcrt

                msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)
                msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)

            self.stdin = sys.stdin
            self.stdout = sys.stdout
            self.write = self._write2
    
    def read(self, sz):
        return self.stdin.read(sz)
    
    def _write2(self, buf):
        self.stdout.write(buf)
        self.stdout.flush()
    
    def _write3(self, buf):
        nw = self.stdout.write(buf)
        left = len(buf) - nw
        while left > 0:
            nw = self.stdout.write(buf[:-left])
            left -= nw
            if nw == 0:
                raise Exception("client lost during write")
        
        self.stdout.flush()


class TcpServer(threading.Thread):
    def __init__(self, ip, port, **kwargs):
        self.srv = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.srv.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.srv.bind((ip, port))
        self.srv.listen(4)
        
        threading.Thread.__init__(self)
        self.daemon = True

    def run(self):
        eprint('listening')
        while True:
            sck, addr = self.srv.accept()
            SftpConnection(FSocket(sck)).start()


class SftpDir(object):
    def __init__(self, path):
        self.path = path
        self.items = os.listdir(path)


class SftpConnection(threading.Thread):
    def __init__(self, f):
        self.f = f
        self.p = Parser(f)
        self.ver = -1
        self.fds = {}
        self.nextfd = 1
        
        threading.Thread.__init__(self)
        self.daemon = True
    
    def _readpkt(self):
        sz = self.p.ru32()
        if sz > 1024 * 1024:
            raise Exception("2big |{:x}| t#{}".format(sz, self.p.u8()))
        
        buf = self.p.read(sz)
        p = Parser(buf)
        itype = p.u8()
        return itype, p
    
    def run(self):
        eprint('conn up')
        
        # rx 01 SSH_FXP_INIT    (just max verison number)
        # tx 02 SSH_FXP_VERSION (with our extension list)
        
        itype, p = self._readpkt()
        if itype != 1:
            raise Exception("expected t#1, got {}".format(itype))
        
        ver = p.ru32()
        self.ver = min(ver, 3)
        eprint('client is ver {}, using {}'.format(ver, self.ver))
        self.f.write(struct.pack(b">IBI", 5, 2, self.ver))
        
        # sshfs makes use of:
        # SFTP_EXT_POSIX_RENAME "posix-rename@openssh.com"
        #   sshfs.ext_posix_rename = 1;
        #   sshfs.rename_workaround = 0;
        # SFTP_EXT_STATVFS "statvfs@openssh.com"
        #   sshfs.ext_statvfs = 1;
        # SFTP_EXT_HARDLINK "hardlink@openssh.com"
        #   sshfs.ext_hardlink = 1;
        # SFTP_EXT_FSYNC "fsync@openssh.com"
        #   sshfs.ext_fsync = 1;
        
        while True:
            itype, p = self._readpkt()
            #eprint("|{:x}| t#{}".format(p.sz - 1, itype))
            # https://tools.ietf.org/html/draft-ietf-secsh-filexfer-02#section-3
            if itype == 7:
                self.rx_lstat(p)
            elif itype == 3:
                self.rx_open(p)
            elif itype == 4:
                self.rx_close(p)
            elif itype == 5:
                self.rx_read(p)
            elif itype == 11:
                self.rx_opendir(p)
            elif itype == 12:
                self.rx_readdir(p)
            else:
                eprint("unhandled", itype)
        
        eprint('conn down')
    
    def _take_handle(self):
        while True:
            # sequential filehandles which rollover at max(uint32)
            if self.nextfd < 0x100:
                hsz = b"B"
            elif self.nextfd < 0x10000:
                hsz = b"H"
            elif self.nextfd < 0x100000000:
                hsz = b"I"
            else:
                # 4 million fds ought to be enough for anybody
                self.nextfd = 1
                continue
            
            handle = struct.pack(hsz, self.nextfd)
            #eprint("pack [{}] [{}] = [{}]".format(hsz, self.nextfd, b2hex(handle)))
            self.nextfd += 1
            if handle not in self.fds:
                return handle
    
    def tx_handle(self, serial, handle):
        # 0x66 102 SSH_FXP_HANDLE
        sz = 9 + len(handle)
        self.f.write(struct.pack(b">IB4sI", sz, 102, serial, len(handle)) + handle)
    
    def rx_open(self, p):
        # 3 SSH_FXP_OPEN
        # https://tools.ietf.org/html/draft-ietf-secsh-filexfer-02#section-6.3
        serial = p.read(4)
        sz = p.ru32()
        fpath = p.read(sz)
        flags = p.ru32()
        attrs = self.dec_attrs(p)
        
        read = flags & 1 << 0
        write = flags & 1 << 1
        append = flags & 1 << 2
        create = flags & 1 << 3
        trunc = flags & 1 << 4
        excl = flags & 1 << 5
        
        # TODO race condition
        if excl and os.path.exists(fpath):
            return self.tx_status(serial, 4)
        
        # https://stackoverflow.com/questions/6648493
        if read and write:
            if append:
                mode = "a+b"
            elif trunc:
                mode = "w+b"
            else:
                mode = "r+b"
        elif read:
            mode = "rb"
        elif write:
            if trunc:
                mode = "wb"
            else:
                mode = "ab"
        else:
            eprint("bad open flags {:x}".format(flags))
            return self.tx_status(serial, 5)
        
        handle = self._take_handle()
        #eprint("using handle |{}| [{}]".format(len(handle), b2hex(handle)))
        try:
            fobj = open(fpath, mode)
            self.fds[handle] = fobj
        except:
            return self.tx_status(serial, 3)  # probably, TODO
        
        self.tx_handle(serial, handle)
    
    def rx_opendir(self, p):
        # 11 SSH_FXP_OPENDIR
        # https://tools.ietf.org/html/draft-ietf-secsh-filexfer-02#section-6.7
        serial = p.read(4)
        sz = p.ru32()
        fpath = p.read(sz)
        try:
            fobj = SftpDir(fpath)
        except Exception as ex:
            eprint("opendir failed,", repr(ex))
            return self.tx_status(serial, 2)
        
        handle = self._take_handle()
        #eprint("using handle |{}| [{}]".format(len(handle), b2hex(handle)))
        self.fds[handle] = fobj
        self.tx_handle(serial, handle)
    
    def rx_readdir(self, p):
        # 12 SSH_FXP_READDIR
        # https://tools.ietf.org/html/draft-ietf-secsh-filexfer-02#section-7
        serial = p.read(4)
        sz = p.ru32()
        handle = p.read(sz)
        try:
            fobj = self.fds[handle]
            fobj.items
        except Exception as ex:
            eprint("readdir failed,", repr(ex))
            return self.tx_status(serial, 2)
        
        n = 0
        ret = b''
        for item in fobj.items:
            n += 1
            ipath = os.path.join(fobj.path, item)
            st = os.stat(ipath)
            attrs = self.enc_attrs(st)
            sz = struct.pack(b">I", len(item))
            ret += sz + item + b"\x00\x00\x00\x00" + attrs
            if len(ret) > 15000:
                break
        
        fobj.items = fobj.items[n:]
        if not ret:
            #eprint("readdir eof")
            return self.tx_status(serial, 1)
        
        # 0x68 104 SSH_FXP_NAME
        # serial + u32(count) + [filename, longname, ATTRS]
        # filename: filename inside folder
        # longname: ls -l
        sz = len(ret) + 9
        self.f.write(struct.pack(b">IB4sI", sz, 104, serial, n) + ret)
    
    def rx_close(self, p):
        # 4 SSH_FXP_CLOSE
        # https://tools.ietf.org/html/draft-ietf-secsh-filexfer-02#section-6.3
        serial = p.read(4)
        sz = p.ru32()
        handle = p.read(sz)
        try:
            fobj = self.fds[handle]
        except:
            eprint("handle {:x} not found".format(struct.unpack(b"I", handle)[0]))
            return self.tx_status(serial, 2)  # probably, TODO
        
        try:
            fobj.close()
        except:
            try:
                fobj.items
            except:
                eprint("close {:x} failed".format(struct.unpack(b"I", handle)[0]))
                return self.tx_status(serial, 4)
        
        del self.fds[handle]
        return self.tx_status(serial, 0)
    
    def rx_read(self, p):
        # 5 SSH_FXP_READ
        # https://tools.ietf.org/html/draft-ietf-secsh-filexfer-02#section-6.4
        serial = p.read(4)
        sz = p.ru32()
        handle = p.read(sz)
        ofs = p.ru64()
        sz = p.ru32()
        try:
            fobj = self.fds[handle]
        except:
            eprint("handle {:x} not found".format(struct.unpack(b"I", handle)[0]))
            return self.tx_status(serial, 2)  # probably, TODO
        
        fobj.seek(ofs)
        npos = fobj.tell()
        if npos != ofs:
            eprint("seek {} failed, {}".format(ofs, npos))
            return self.tx_status(serial, 5)  # not in spec but sounds correct
        
        buf = fobj.read(sz)
        if not buf:
            return self.tx_status(serial, 1)
        
        # 103 SSH_FXP_DATA
        # https://tools.ietf.org/html/draft-ietf-secsh-filexfer-02#section-7
        self.f.write(struct.pack(b">IB4sI",
            len(buf) + 9, 103, serial, len(buf)) + buf)
        
    def rx_lstat(self, p):
        # 7 SSH_FXP_LSTAT
        # https://tools.ietf.org/html/draft-ietf-secsh-filexfer-02#section-6.8
        serial = p.read(4)
        sz = p.ru32()
        fpath = p.read(sz)
        try:
            st = os.lstat(fpath)
            self.tx_attrs(serial, st)
        except IOError as ex:
            eprint(repr(ex))
            self.tx_status(serial, 2)  # TODO be specific
    
    def dec_attrs(self, p):
        # https://tools.ietf.org/html/draft-ietf-secsh-filexfer-02#section-5
        #eprint("TODO: ATTRS are dicsarded")
        pass
    
    def enc_attrs(self, st):
        return struct.pack(b">IQIIIII", 15,
            st.st_size, st.st_uid, st.st_gid, st.st_mode,
            int(st.st_atime), int(st.st_mtime))
    
    def tx_attrs(self, serial, st):
        # 105 SSH_FXP_ATTRS
        ret = self.enc_attrs(st)
        ret = struct.pack(b">IB4s", len(ret) + 5, 105, serial) + ret
        self.f.write(ret)
    
    def tx_status(self, serial, code):
        # 0x65 101 SSH_FXP_STATUS
        # https://tools.ietf.org/html/draft-ietf-secsh-filexfer-02#section-7
        # 0ok 1eof 2enoent 3eperm 4fail 5badmsg 8unsupp
        self.f.write(struct.pack(b">IB4sIIcI",
            18, 101, serial, code, 1, b"h", 0))


class APFormatter(
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.RawDescriptionHelpFormatter):
    pass


if __name__ == '__main__':
    ap = argparse.ArgumentParser(
        formatter_class=APFormatter,
        epilog="""
examples:
  foo
""")

    ap.add_argument("-l", action="store_true", help='use tcp instead of stdin/out')
    ap.add_argument("-i", metavar='IP', type=str, help='IP to bind', default='127.0.0.1')
    ap.add_argument("-p", metavar='PORT', type=int, help='port to listen on', default=5289)
    ar = ap.parse_args()

    if ar.l:
        srv = TcpServer(ar.i, ar.p)
        srv.start()
    else:
        conn = SftpConnection(FStdio())
        conn.start()

    while True:
        time.sleep(9001)


# tcp debug:
# f=/dev/shm/sshfs.fifo; rm -f $f; mkfifo -m600 $f; socat -x -v tcp4:127.0.0.1:5289 - <$f | sshfs -o slave :/dev/shm/r/src/d /dev/shm/r/fuse/d >$f

# tcp:
# f=/dev/shm/sshfs.fifo; rm -f $f; mkfifo -m600 $f; socat tcp4:127.0.0.1:5289 - <$f | sshfs -o slave :/dev/shm/r/src/d /dev/shm/r/fuse/d >$f

# ssh:
# f=/dev/shm/sshfs.fifo; rm -f $f; mkfifo -m600 $f; ./sftp-srv.py <$f | ssh -p 573 127.0.0.1 sshfs -o slave :/dev/shm/r/src/d /dev/shm/r/fuse/d >$f

# stdio debug:
# f=/dev/shm/sshfs.fifo; rm -f $f; mkfifo -m600 $f; socat -t100 -x -v - - <$f 2> >(sed $'s/^/\033[33m/;s/$/\033[0m/' >&2) | ./sftp-srv.py | socat -t100 -x -v - - 2> >(sed $'s/^/\033[36m/;s/$/\033[0m/' >&2) | sshfs -o slave :/dev/shm/r/src/d /dev/shm/r/fuse/d >$f

# stdio:
# f=/dev/shm/sshfs.fifo; rm -f $f; mkfifo -m600 $f; ./sftp-srv.py <$f | sshfs -o slave :/dev/shm/r/src/d /dev/shm/r/fuse/d >$f

# stresstest:
# t=/dev/shm/r/src/d; rm -rf $t; mkdir -p $t; (cd $t && for a in q w e r t y u i o p; do echo $a; for b in a s d f g h j k l; do for c in z x c v b n m; do for d in 1 2 3 4 5 6 7 8 9; do y=$a/$b/$c/$d; z=$a$b$c$d; mkdir -p $y; echo $z-foo > $y/$z; done; done; done; done); tar -cC/ $t | pigz -c > /dev/shm/src.tgz
# mkdir -p /dev/shm/r/fuse/d && cd /dev/shm &&
#   (cd /dev/shm/r/src/ && hashwalk x) >mdw.src
# f=/dev/shm/sshfs.fifo; rm -f $f; mkfifo -m600 $f; ~/dev/sftp-srv.py <$f | sshfs -o slave :/dev/shm/r/src/d /dev/shm/r/fuse/d >$f
#   (cd /dev/shm/r/fuse/ && hashwalk x) >mdw.sftp-srv
